MAX_PERCENTS = 100.0


class Bar(object):
    bars = None

    def __init__(self, value):
        self.value = value


class HBar(Bar):
    bars = [
        "\u2581",
        "\u2582",
        "\u2583",
        "\u2584",
        "\u2585",
        "\u2586",
        "\u2587",
        "\u2588",
    ]

    """This class is a helper class used to draw horizontal bars - please use hbar directly

    :param value: percentage value to draw (float, between 0 and 100)
    """

    def __init__(self, value):
        super(HBar, self).__init__(value)
        self.step = MAX_PERCENTS / len(HBar.bars)

    def get_char(self):
        """Returns the character representing the current object's value

        :return: character representing the value passed during initialization
        :rtype: string with one character
        """
        for i in range(len(HBar.bars)):
            left = i * self.step
            right = (i + 1) * self.step
            if left <= self.value < right:
                return self.bars[i]
        return self.bars[-1]


def hbar(value):
    """"Retrieves the horizontal bar character representing the input value

    :param value: percentage value to draw (float, between 0 and 100)
    :return: character representing the value passed during initialization
    :rtype: string with one character
    """
    return HBar(value).get_char()


class VBar(Bar):
    bars = [
        "\u258f",
        "\u258e",
        "\u258d",
        "\u258c",
        "\u258b",
        "\u258a",
        "\u2589",
        "\u2588",
    ]

    """This class is a helper class used to draw vertical bars - please use vbar directly

    :param value: percentage value to draw (float, between 0 and 100)
    :param width: maximum width of the bar in characters
    """

    def __init__(self, value, width=1):
        super(VBar, self).__init__(value)
        self.step = MAX_PERCENTS / (len(VBar.bars) * width)
        self.width = width

    """Returns the characters representing the current object's value

    :return: characters representing the value passed during initialization
    :rtype: string
    """

    def get_chars(self):
        if self.value == 100:
            return self.bars[-1] * self.width
        if self.width == 1:
            for i in range(len(VBar.bars)):
                left = i * self.step
                right = (i + 1) * self.step
                if left <= self.value < right:
                    return self.bars[i]
        else:
            full_parts = int(self.value // (self.step * len(Vbar.bars)))
            remainder = self.value - full_parts * self.step * CHARS
            empty_parts = self.width - full_parts
            if remainder >= 0:
                empty_parts -= 1
            part_vbar = VBar(remainder * self.width)  # scale to width
            chars = self.bars[-1] * full_parts
            chars += part_vbar.get_chars()
            chars += " " * empty_parts
            return chars


def vbar(value, width):
    """Returns the characters representing the current object's value

    :param value: percentage value to draw (float, between 0 and 100)
    :param width: maximum width of the bar in characters

    :return: characters representing the value passed during initialization
    :rtype: string
    """
    return VBar(value, width).get_chars()


class BrailleGraph(object):
    chars = {
        (0, 0): " ",
        (1, 0): "\u2840",
        (2, 0): "\u2844",
        (3, 0): "\u2846",
        (4, 0): "\u2847",
        (0, 1): "\u2880",
        (0, 2): "\u28a0",
        (0, 3): "\u28b0",
        (0, 4): "\u28b8",
        (1, 1): "\u28c0",
        (2, 1): "\u28c4",
        (3, 1): "\u28c6",
        (4, 1): "\u28c7",
        (1, 2): "\u28e0",
        (2, 2): "\u28e4",
        (3, 2): "\u28e6",
        (4, 2): "\u28e7",
        (1, 3): "\u28f0",
        (2, 3): "\u28f4",
        (3, 3): "\u28f6",
        (4, 3): "\u28f7",
        (1, 4): "\u28f8",
        (2, 4): "\u28fc",
        (3, 4): "\u28fe",
        (4, 4): "\u28ff",
    }

    """This class is a helper class used to draw braille graphs - please use braille directly

    :param values: values to draw
    """

    def __init__(self, values):
        self.values = values
        # length of values list must be even
        # because one Braille char displays two values
        if len(self.values) % 2 == 1:
            self.values.append(0)
        self.steps = self.get_steps()
        self.parts = [tuple(self.steps[i : i + 2]) for i in range(len(self.steps))[::2]]

    @staticmethod
    def get_height(value, unit):
        if value < unit / 10.0:
            return 0
        elif value <= unit:
            return 1
        elif value <= unit * 2:
            return 2
        elif value <= unit * 3:
            return 3
        else:
            return 4

    def get_steps(self):
        maxval = max(self.values)
        unit = maxval / 4.0
        if unit == 0:
            return [0] * len(self.values)
        stepslist = []
        for value in self.values:
            stepslist.append(self.get_height(value, unit))
        return stepslist

    def get_chars(self):
        chars = []
        for part in self.parts:
            chars.append(BrailleGraph.chars[part])
        return "".join(chars)


def braille(values):
    return BrailleGraph(values).get_chars()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the status of watson (time-tracking tool)

Requires the following executable:
    * watson

contributed by `bendardenne <https://github.com/bendardenne>`_ - many thanks!
"""

import logging
import re
import functools

import core.module
import core.widget
import core.input
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.text))

        self.__tracking = False
        self.__project = ""
        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.toggle)

    def toggle(self, widget):
        self.__project = "hit"
        if self.__tracking:
            util.cli.execute("watson stop")
        else:
            util.cli.execute("watson restart")
        self.__tracking = not self.__tracking

    def text(self, widget):
        if self.__tracking:
            return self.__project
        else:
            return "Paused"

    def update(self):
        output = util.cli.execute("watson status")
        if re.match(r"No project started", output):
            self.__tracking = False
            return

        self.__tracking = True
        m = re.search(r"Project (.+) started", output)
        self.__project = m.group(1)

    def state(self, widget):
        return "on" if self.__tracking else "off"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Retrieves location information from an external
service and caches it for 12h (retries are done every
30m in case of problems)

Right now, it uses (in order of preference):
    - http://free.ipwhois.io/
    - http://ipapi.co/
"""


import json
import time
import urllib.request

__document = None
__data = {}
__next = 0
__sources = [
    {
        "url": "http://ipapi.co/json",
        "mapping": {
            "latitude": "latitude",
            "longitude": "longitude",
            "country_name": "country",
            "ip": "public_ip",
        },
    },
    {
        "url": "http://free.ipwhois.io/json/",
        "mapping": {
            "latitude": "latitude",
            "longitude": "longitude",
            "country": "country",
            "ip": "public_ip",
        },
    }
]


def __expired():
    global __next
    return __next <= time.time()


def __load():
    global __data
    global __next

    __data = {}
    for src in __sources:
        try:
            tmp = json.loads(urllib.request.urlopen(src["url"]).read())
            for k, v in src["mapping"].items():
                __data[v] = tmp.get(k, None)
            __next = time.time() + 60 * 60 * 12  # update once every 12h
            return
        except Exception as e:
            pass
    __next = time.time() + 60 * 30  # error - try again every 30m


def __get(name, default=None):
    global __data
    if not __data or __expired():
        __load()
    return __data.get(name, default)


def reset():
    """Resets the location library, ensuring that a new query will be started
    """
    global __next
    __next = 0


def coordinates():
    """Returns a latitude, longitude pair

    :return: current latitude and longitude
    :rtype: pair of strings
    """
    return __get("latitude"), __get("longitude")


def country():
    """Returns the current country name

    :return: country name
    :rtype: string
    """
    return __get("country")


def public_ip():
    """Returns the current public IP

    :return: public IP
    :rtype: string
    """
    return __get("public_ip")


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Check updates to Arch Linux.

Requires the following executable:
    * checkupdates (from pacman-contrib)

contributed by `lucassouto <https://github.com/lucassouto>`_ - many thanks!
"""

import logging

import core.module
import core.widget
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.utilization))
        self.background = True
        self.__packages = 0
        self.__error = False

    @property
    def __format(self):
        return self.parameter("format", "Update Arch: {}")

    def utilization(self, widget):
        return self.__format.format(self.__packages)

    def hidden(self):
        return self.__packages == 0 and not self.__error

    def update(self):
        self.__error = False
        code, result = util.cli.execute(
            "checkupdates", ignore_errors=True, return_exitcode=True
        )

        if code == 0:
            self.__packages = len(result.split("\n"))
        elif code == 2:
            self.__packages = 0
        else:
            self.__error = True
            logging.error("checkupdates exited with {}: {}".format(code, result))

    def state(self, widget):
        if self.__error:
            return "warning"
        return self.threshold_state(self.__packages, 1, 100)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Display and run a Pomodoro timer.
Left click to start timer, left click again to pause.
Right click will cancel the timer.

Parameters:
    * pomodoro.work: The work duration of timer in minutes (defaults to 25)
    * pomodoro.break: The break duration of timer in minutes (defaults to 5)
    * pomodoro.format: Timer display format with '%m' and '%s' for minutes and seconds (defaults to '%m:%s')
      Examples: '%m min %s sec', '%mm', '', 'timer'
    * pomodoro.notify: Notification command to run when timer ends/starts (defaults to nothing)
      Example: 'notify-send 'Time up!''. If you want to chain multiple commands,
      please use an external wrapper script and invoke that. The module itself does
      not support command chaining (see https://github.com/tobi-wan-kenobi/bumblebee-status/issues/532
      for a detailled explanation)

contributed by `martindoublem <https://github.com/martindoublem>`_, inspired by `karthink <https://github.com/karthink>`_ - many thanks!
"""

from __future__ import absolute_import
import datetime
from math import ceil

import core.module
import core.widget
import core.input

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.text))

        # Parameters
        self.__work_period = int(self.parameter("work", 25))
        self.__break_period = int(self.parameter("break", 5))
        self.__time_format = self.parameter("format", "%m:%s")
        self.__notify_cmd = self.parameter("notify", "")

        # TODO: Handle time formats more gracefully. This is kludge.
        self.display_seconds_p = False
        self.display_minutes_p = False
        if "%s" in self.__time_format:
            self.display_seconds_p = True
        if "%m" in self.__time_format:
            self.display_minutes_p = True

        self.remaining_time = datetime.timedelta(minutes=self.__work_period)

        self.time = None
        self.pomodoro = {"state": "OFF", "type": ""}
        self.__text = self.remaining_time_str() + self.pomodoro["type"]

        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd=self.timer_play_pause
        )
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self.timer_reset)

    def remaining_time_str(self):
        if self.display_seconds_p and self.display_minutes_p:
            minutes, seconds = divmod(self.remaining_time.seconds, 60)
        if not self.display_seconds_p:
            minutes = ceil(self.remaining_time.seconds / 60)
            seconds = 0
        if not self.display_minutes_p:
            minutes = 0
            seconds = self.remaining_time.seconds

        minutes = "{:2d}".format(minutes)
        seconds = "{:02d}".format(seconds)
        return self.__time_format.replace("%m", minutes).replace("%s", seconds) + " "

    def text(self, widget):
        return "{}".format(self.__text)

    def update(self):
        if self.pomodoro["state"] == "ON":
            timediff = datetime.datetime.now() - self.time
            if timediff.seconds >= 0:
                self.remaining_time -= timediff
                self.time = datetime.datetime.now()

            if self.remaining_time.total_seconds() <= 0:
                self.notify()
                if self.pomodoro["type"] == "Work":
                    self.pomodoro["type"] = "Break"
                    self.remaining_time = datetime.timedelta(
                        minutes=self.__break_period
                    )
                elif self.pomodoro["type"] == "Break":
                    self.pomodoro["type"] = "Work"
                    self.remaining_time = datetime.timedelta(minutes=self.__work_period)

        self.__text = self.remaining_time_str() + self.pomodoro["type"]

    def notify(self):
        if self.__notify_cmd:
            util.cli.execute(self.__notify_cmd)

    def timer_play_pause(self, widget):
        if self.pomodoro["state"] == "OFF":
            self.pomodoro = {"state": "ON", "type": "Work"}
            self.remaining_time = datetime.timedelta(minutes=self.__work_period)
            self.time = datetime.datetime.now()
        elif self.pomodoro["state"] == "ON":
            self.pomodoro["state"] = "PAUSED"
            self.remaining_time -= datetime.datetime.now() - self.time
            self.time = datetime.datetime.now()
        elif self.pomodoro["state"] == "PAUSED":
            self.pomodoro["state"] = "ON"
            self.time = datetime.datetime.now()

    def timer_reset(self, widget):
        if self.pomodoro["state"] == "ON" or self.pomodoro["state"] == "PAUSED":
            self.pomodoro = {"state": "OFF", "type": ""}
            self.remaining_time = datetime.timedelta(minutes=self.__work_period)

    def state(self, widget):
        state = []
        state.append(self.pomodoro["state"].lower())
        if self.pomodoro["state"] == "ON" or self.pomodoro["state"] == "OFF":
            state.append(self.pomodoro["type"].lower())

        return state


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the name, IP address(es) and status of each available network interface.

Requires the following python module:
    * netifaces

Parameters:
    * nic.exclude: Comma-separated list of interface prefixes to exclude (defaults to 'lo,virbr,docker,vboxnet,veth,br')
    * nic.include: Comma-separated list of interfaces to include
    * nic.states: Comma-separated list of states to show (prefix with '^' to invert - i.e. ^down -> show all devices that are not in state down)
    * nic.format: Format string (defaults to '{intf} {state} {ip} {ssid}')
"""

import shutil
import netifaces
import subprocess

import core.module
import core.decorators
import util.cli
import util.format


class Module(core.module.Module):
    @core.decorators.every(seconds=10)
    def __init__(self, config, theme):
        widgets = []
        super().__init__(config, theme, widgets)
        self._exclude = tuple(
            filter(
                len,
                self.parameter("exclude", "lo,virbr,docker,vboxnet,veth,br").split(","),
            )
        )
        self._include = self.parameter("include", "").split(",")

        self._states = {"include": [], "exclude": []}
        for state in tuple(
            filter(len, util.format.aslist(self.parameter("states", "")))
        ):
            if state[0] == "^":
                self._states["exclude"].append(state[1:])
            else:
                self._states["include"].append(state)
        self._format = self.parameter("format", "{intf} {state} {ip} {ssid}")
        self.iwgetid = shutil.which("iwgetid")
        self._update_widgets(widgets)

    def update(self):
        self._update_widgets(self.widgets())

    def state(self, widget):
        states = []

        if widget.get("state") == "down":
            states.append("critical")
        elif widget.get("state") != "up":
            states.append("warning")

        intf = widget.get("intf")
        iftype = "wireless" if self._iswlan(intf) else "wired"
        iftype = "tunnel" if self._istunnel(intf) else iftype

        states.append("{}-{}".format(iftype, widget.get("state")))

        return states

    def _iswlan(self, intf):
        # wifi, wlan, wlp, seems to work for me
        if intf.startswith("w"):
            return True
        return False

    def _istunnel(self, intf):
        return intf.startswith("tun") or intf.startswith("wg")

    def get_addresses(self, intf):
        retval = []
        try:
            for ip in netifaces.ifaddresses(intf).get(netifaces.AF_INET, []):
                if ip.get("addr", "") != "":
                    retval.append(ip.get("addr"))
        except Exception:
            return []
        return retval

    def _update_widgets(self, widgets):
        self.clear_widgets()
        interfaces = [
            i for i in netifaces.interfaces() if not i.startswith(self._exclude)
        ]
        interfaces.extend([i for i in netifaces.interfaces() if i in self._include])

        for intf in interfaces:
            addr = []
            state = "down"
            for ip in self.get_addresses(intf):
                addr.append(ip)
                state = "up"

            if len(self._states["exclude"]) > 0 and state in self._states["exclude"]:
                continue
            if (
                len(self._states["include"]) > 0
                and state not in self._states["include"]
            ):
                continue

            widget = self.widget(intf)
            if not widget:
                widget = self.add_widget(name=intf)
            # join/split is used to get rid of multiple whitespaces (in case SSID is not available, for instance
            widget.full_text(
                " ".join(
                    self._format.format(
                        ip=", ".join(addr),
                        intf=intf,
                        state=state,
                        ssid=self.get_ssid(intf),
                    ).split()
                )
            )
            widget.set("intf", intf)
            widget.set("state", state)

    def get_ssid(self, intf):
        if self._iswlan(intf) and not self._istunnel(intf) and self.iwgetid:
            return util.cli.execute(
                "{} -r {}".format(self.iwgetid, intf), ignore_errors=True
            )
        return ""


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""
Displays the unread GitHub notifications count for a GitHub user using the following reasons:

    * https://developer.github.com/v3/activity/notifications/#notification-reasons

Requires the following library:
    * requests

Parameters:
    * github.token: GitHub user access token, the token needs to have the 'notifications' scope.
    * github.interval: Interval in minutes between updates, default is 5.
    * github.reasons: Comma separated reasons to be parsed (e.g.: github.reasons=mention,team_mention,review_requested)

contributed by:
    * v1 - `yvesh <https://github.com/yvesh>`_ - many thanks!
    * v2 - `cristianmiranda <https://github.com/cristianmiranda>`_ - many thanks!
"""

import shutil
import requests

import core.module
import core.widget
import core.decorators
import core.input

import util.format


class Module(core.module.Module):
    @core.decorators.every(minutes=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.github))

        self.background = True
        self.__count = 0
        self.__label = ""
        self.__requests = requests.Session()
        self.__requests.headers.update(
            {"Authorization": "token {}".format(self.parameter("token", ""))}
        )

        self.__reasons = []
        reasons = self.parameter("reasons", "")
        if reasons:
            self.__reasons = util.format.aslist(reasons)

        cmd = "xdg-open"
        if not shutil.which(cmd):
            cmd = "x-www-browser"

        core.input.register(
            self,
            button=core.input.LEFT_MOUSE,
            cmd="{} https://github.com/notifications".format(cmd),
        )

    def github(self, _):
        return str(self.__label)

    def update(self):
        try:
            url = "https://api.github.com/notifications"
            notifications = self.__requests.get(url)

            total = self.__getTotalUnreadNotificationsCount(notifications)
            self.__count = total
            self.__label = str(total)

            counts = []
            if len(self.__reasons) > 0:
                for reason in self.__reasons:
                    unread = self.__getUnreadNotificationsCountByReason(
                        notifications, reason
                    )
                    counts.append(str(unread))

                self.__label += " - "
                self.__label += "/".join(counts)

        except Exception as err:
            print(err)
            self.__label = "n/a"

    def __getUnreadNotificationsCountByReason(self, notifications, reason):
        return len(
            list(
                filter(
                    lambda notification: notification["unread"]
                    and notification["reason"] == reason,
                    notifications.json(),
                )
            )
        )

    def __getTotalUnreadNotificationsCount(self, notifications):
        return len(
            list(
                filter(
                    lambda notification: notification["unread"], notifications.json()
                )
            )
        )

    def state(self, widget):
        state = []

        if self.__count > 0:
            state.append("warning")

        return state


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the system hostname.

contributed by `varkokonyi <https://github.com/varkokonyi>`_ - many thanks!
"""

import platform

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))
        self.__hname = ""

    def output(self, _):
        return self.__hname + " " + "\uf233"

    def update(self):
        self.__hname = platform.node()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python
"""Setup file for bumbleestatus bar to allow pip install of full package"""
# -*- coding: utf8 - *-
from setuptools import setup
import versioneer

with open("requirements/base.txt") as f:
    INSTALL_REQS = [line for line in f.read().split("\n") if line]

# Module packages
def read_module(filename):
    """Read each in a module's requirements and parse it for extras"""
    with open("requirements/modules/{}.txt".format(filename)) as fname:
        return [rline for rline in fname.read().split("\n") if rline]


EXTRAS_REQUIREMENTS_MAP = {
    "battery-upower": read_module("battery_upower_reqs"),
    "cpu": read_module("cpu"),
    "cpu2": read_module("cpu2"),
    "currency": read_module("currency"),
    "docker_ps": read_module("docker_ps"),
    "dunst": read_module("dunst"),
    "getcrypto": read_module("getcrypto"),
    "git": read_module("git"),
    "github": read_module("github"),
    "hddtemp": read_module("hddtemp"),
    "layout-xkb": read_module("layout_xkb"),
    "memory": read_module("memory"),
    "network_traffic": read_module("network_traffic"),
    "nic": read_module("nic"),
    "pihole": read_module("pihole"),
    "rss": read_module("rss"),
    "spaceapi": read_module("spaceapi"),
    "spotify": read_module("spotify"),
    "stock": read_module("stock"),
    "sun": read_module("sun"),
    "system": read_module("system"),
    "taskwarrior": read_module("taskwarrior"),
    "title": read_module("title"),
    "traffic": read_module("traffic"),
    "weather": read_module("weather"),
    "yubikey": read_module("yubikey"),
}

import glob

setup(
    #    packages=["bumblebee-status-packages"],
    #    package_dir={"bumblebee-status-packages": "."},
    install_requires=INSTALL_REQS,
    extras_require=EXTRAS_REQUIREMENTS_MAP,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
    test_suite="tests",
    data_files=[
        ("share/bumblebee-status/themes", glob.glob("themes/*.json")),
        ("share/bumblebee-status/themes/icons", glob.glob("themes/icons/*.json")),
        ("share/bumblebee-status/utility", glob.glob("bin/*")),
    ],
)



import difflib
import logging

import util.format

log = logging.getLogger(__name__)


"""Specifies that a module should never update (i.e. has static content).
This means that its update() method will never be invoked

:param init: The __init__() method of the module

:return: Wrapped method that sets the module's interval to "never"
"""


def never(init):
    def call_init(obj, *args, **kwargs):
        init(obj, *args, **kwargs)
        if obj.parameter("interval") is None:
            obj.set("interval", "never")

    return call_init


"""Specifies the interval for executing the module's update() method

:param hours: Hours between two update() invocations, defaults to 0
:param minutes: Minutes between two update() invocations, defaults to 0
:param seconds: Seconds between two update() invocations, defaults to 0

:return: Wrapped method that sets the module's interval correspondingly
"""


def every(hours=0, minutes=0, seconds=0):
    def decorator_init(init):
        def call_init(obj, *args, **kwargs):
            init(obj, *args, **kwargs)
            if obj.parameter("interval") is None:
                obj.set("interval", hours * 3600 + minutes * 60 + seconds)

        return call_init

    return decorator_init


"""Specifies that the module's content should scroll, if required

The exact behaviour of this method is governed by a number of parameters,
specifically: The module's parameter "scrolling.width"  specifies the width when
scrolling starts, "scrolling.makewide" defines whether the module should be expanded
to "scrolling.width" automatically, if the content is shorter, the parameter
"scrolling.bounce" defines whether it scrolls like a marquee (False) or should bounce
when the end of the content is reached. "scrolling.speed" defines the number of characters
to scroll each iteration.

:param func: Function for which the result should be scrolled
"""


def scrollable(func):
    def wrapper(module, widget):
        text = func(module, widget)
        if not text:
            return text

        if (
            difflib.SequenceMatcher(a=text, b=widget.get("__content__", text)).ratio()
            < 0.9
        ):
            widget.set("scrolling.start", 0)
            widget.set("scrolling.direction", "right")
        widget.set("__content__", text)

        width = util.format.asint(module.parameter("scrolling.width", 30))
        if util.format.asbool(module.parameter("scrolling.makewide", True)):
            widget.set("theme.minwidth", "A" * width)
        if width < 0 or len(text) <= width:
            return text

        start = widget.get("scrolling.start", 0)
        bounce = util.format.asbool(module.parameter("scrolling.bounce", True))
        scroll_speed = util.format.asint(module.parameter("scrolling.speed", 1))
        direction = widget.get("scrolling.direction", "right")

        if direction == "left":
            if start - scroll_speed < 0:  # bounce back
                widget.set("scrolling.direction", "right")
            else:
                scroll_speed = -scroll_speed

        next_start = start + scroll_speed
        if next_start + width > len(text):
            if not bounce:
                next_start = 0
            else:
                next_start = start - scroll_speed
                widget.set("scrolling.direction", "left")

        widget.set("scrolling.start", next_start)

        return text[start : start + width]

    return wrapper


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current date and time.

Parameters:
    * datetime.format: strftime()-compatible formatting string
    * datetime.locale: locale to use rather than the system default
"""

from __future__ import absolute_import
import datetime
import locale

import core.module
import core.widget
import core.input


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.full_text))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd="calendar")
        self._fmt = self.parameter("format", self.default_format())
        l = locale.getdefaultlocale()
        if not l or l == (None, None):
            l = ("en_US", "UTF-8")
        lcl = self.parameter("locale", ".".join(l))
        try:
            locale.setlocale(locale.LC_TIME, lcl.split("."))
        except Exception as e:
            locale.setlocale(locale.LC_TIME, ("en_US", "UTF-8"))

    def default_format(self):
        return "%x %X"

    def full_text(self, widget):
        enc = locale.getpreferredencoding()
        retval = datetime.datetime.now().strftime(self._fmt)
        if hasattr(retval, "decode"):
            return retval.decode(enc)
        return retval


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current date and time with timezone options.

Parameters:
    * datetimetz.format   : strftime()-compatible formatting string
    * datetimetz.timezone : IANA timezone name
    * datetz.format       : alias for datetimetz.format
    * timetz.format       : alias for datetimetz.format
    * timetz.timezone     : alias for datetimetz.timezone
    * datetimetz.locale   : locale to use rather than the system default
    * datetz.locale       : alias for datetimetz.locale
    * timetz.locale       : alias for datetimetz.locale
    * timetz.timezone     : alias for datetimetz.timezone

contributed by `frankzhao <https://github.com/frankzhao>`_ - many thanks!
"""

from __future__ import absolute_import
import datetime
import locale
import logging
import pytz
import tzlocal

import core.module
import core.widget
import core.input

import util.format


def default_format(module):
    default = "%x %X %Z"
    if module == "datetz":
        default = "%x %Z"
    if module == "timetz":
        default = "%X %Z"
    return default


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.get_time))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.next_tz)
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self.prev_tz)
        self.__fmt = self.parameter("format", self.default_format())

        default_timezone = ""
        try:
            default_timezone = tzlocal.get_localzone().zone
        except Exception as e:
            logging.error("unable to get default timezone: {}".format(str(e)))
        try:
            self._timezones = util.format.aslist(
                self.parameter("timezone", default_timezone)
            )
        except:
            self._timezones = [default_timezone]
        self._current_tz = 0

        l = locale.getdefaultlocale()
        if not l or l == (None, None):
            l = ("en_US", "UTF-8")
        lcl = self.parameter("locale", ".".join(l))
        try:
            locale.setlocale(locale.LC_TIME, lcl.split("."))
        except Exception:
            locale.setlocale(locale.LC_TIME, ("en_US", "UTF-8"))

    def default_format(self):
        return "%x %X %Z"

    def get_time(self, widget):
        try:
            try:
                tz = pytz.timezone(self._timezones[self._current_tz].strip())
                retval = (
                    datetime.datetime.now(tz=tzlocal.get_localzone())
                    .astimezone(tz)
                    .strftime(self.__fmt)
                )
            except pytz.exceptions.UnknownTimeZoneError:
                retval = "[Unknown timezone: {}]".format(
                    self._timezones[self._current_tz].strip()
                )
        except Exception as e:
            logging.error("unable to get time: {}".format(str(e)))
            retval = "[n/a]"

        enc = locale.getpreferredencoding()
        if hasattr(retval, "decode"):
            return retval.decode(enc)
        return retval

    def next_tz(self, event):
        next_timezone = self._current_tz + 1
        if next_timezone >= len(self._timezones):
            next_timezone = 0  # wraparound
        self._current_tz = next_timezone

    def prev_tz(self, event):
        previous_timezone = self._current_tz - 1
        if previous_timezone < 0:
            previous_timezone = 0  # wraparound
        self._current_tz = previous_timezone


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0112,R0903

"""Shows a widget per user-defined shortcut and allows to define the behaviour
when clicking on it.

For more than one shortcut, the commands and labels are strings separated by
a demiliter (; semicolon by default).

For example in order to create two shortcuts labeled A and B with commands
cmdA and cmdB you could do:

 ./bumblebee-status -m shortcut -p shortcut.cmd='ls;ps' shortcut.label='A;B'

Parameters:
    * shortcut.cmds  : List of commands to execute
    * shortcut.labels: List of widgets' labels (text)
    * shortcut.delim : Commands and labels delimiter (; semicolon by default)


contributed by `cacyss0807 <https://github.com/cacyss0807>`_ - many thanks!
"""

import logging

LINK = "https://github.com/tobi-wan-kenobi/bumblebee-status/wiki"
LABEL = "Click me"

import core.module
import core.input
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.__labels = self.parameter("labels", "{}".format(LABEL))
        self.__cmds = self.parameter("cmds", "firefox {}".format(LINK))
        self.__delim = self.parameter("delim", ";")

        self.update_widgets()

    def update_widgets(self):
        """ Creates a set of widget per user define shortcut."""

        cmds = self.__cmds.split(self.__delim)
        labels = self.__labels.split(self.__delim)

        # to be on the safe side create as many widgets as there are data (cmds or labels)
        num_shortcuts = min(len(cmds), len(labels))

        # report possible problem as a warning
        if len(cmds) is not len(labels):
            logging.warning(
                "shortcut: the number of commands does not match "
                "the number of provided labels."
            )
            logging.warning("cmds : %s, labels : %s", cmds, labels)

        for idx in range(0, num_shortcuts):
            cmd = cmds[idx]
            label = labels[idx]

            widget = self.add_widget(full_text=label)
            core.input.register(widget, button=core.input.LEFT_MOUSE, cmd=cmd)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Copy passwords from a password store into the clipboard (currently supports only 'pass')

Many thanks to [@bbernhard](https://github.com/bbernhard) for the idea!

Parameters:
    * vault.duration: Duration until password is cleared from clipboard (defaults to 30)
    * vault.location: Location of the password store (defaults to ~/.password-store)
    * vault.offx: x-axis offset of popup menu (defaults to 0)
    * vault.offy: y-axis offset of popup menu (defaults to 0)

Many thanks to `bbernhard <https://github.com/bbernhard>`_ for the idea!
"""


# TODO:
# - support multiple backends by abstracting the menu structure into a tree
# - build the menu and the actions based on that abstracted tree
#

import os
import time
import threading

import core.module
import core.widget
import core.input
import core.event

import util.cli
import util.popup


def generate_callback(callback, path, name):
    return lambda: callback(os.path.join(path, name))

def build_menu(parent, current_directory, callback):
    with os.scandir(current_directory) as it:
        for entry in it:
            if entry.name.startswith("."):
                continue
            if entry.is_file():
                name = entry.name[: entry.name.rfind(".")]
                parent.add_menuitem(
                    name,
                    callback=generate_callback(callback, current_directory, name),
                )

            else:
                submenu = util.popup.menu(parent, leave=False)
                build_menu(
                    submenu, os.path.join(current_directory, entry.name), callback
                )
                parent.add_cascade(entry.name, submenu)


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.text))

        self.__duration = int(self.parameter("duration", 30))
        self.__offx = int(self.parameter("offx", 0))
        self.__offy = int(self.parameter("offy", 0))
        self.__path = os.path.expanduser(
            self.parameter("location", "~/.password-store/")
        )
        self.__reset()
        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.popup)

    def popup(self, widget):
        menu = util.popup.menu(leave=False)

        build_menu(menu, self.__path, self.__callback)
        menu.show(widget, offset_x=self.__offx, offset_y=self.__offy)

    def __reset(self):
        self.__timer = None
        self.__text = str(self.parameter("text", "<click-for-password>"))

    def __callback(self, secret_name):
        secret_name = secret_name.replace(self.__path, "")  # remove common path
        if self.__timer:
            self.__timer.cancel()
        env = os.environ
        env["PASSWORD_STORE_CLIP_TIME"] = str(self.__duration)
        res = util.cli.execute(
            "pass show -c {}".format(secret_name),
            wait=False,
            env=env,
            ignore_errors=True,
        )
        self.__timer = threading.Timer(self.__duration, self.__reset)
        self.__timer.start()
        self.__start = int(time.time())
        self.__text = secret_name

    def text(self, widget):
        if self.__timer:
            return "{} ({}s)".format(
                self.__text, self.__duration - (int(time.time()) - self.__start)
            )
        return self.__text


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: utf-8 -*-

"""Fetch hard drive temeperature data from a hddtemp daemon
that runs on localhost and default port (7634)

contributed by `somospocos <https://github.com/somospocos>`_ - many thanks!
"""

import socket

import core.module
import core.widget

HOST = "localhost"
PORT = 7634

CHUNK_SIZE = 1024
RECORD_SIZE = 5
SEPARATOR = "|"


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.hddtemps))
        self.__hddtemps = self.__get_hddtemps()

    def hddtemps(self, _):
        return self.__hddtemps

    def __fetch_data(self):
        """fetch data from hddtemp service"""
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.connect((HOST, PORT))
                data = ""
                while True:
                    chunk = sock.recv(CHUNK_SIZE)
                    if chunk:
                        data += str(chunk)
                    else:
                        break
            return data
        except (AttributeError, socket.error) as e:
            pass

    @staticmethod
    def __get_parts(data):
        """
            split data using | separator and remove first item
            (because the first item is empty)
        """
        parts = data.split("|")[1:]
        return parts

    @staticmethod
    def __partition_parts(parts):
        """
            partition parts: one device record is five (5) items
        """
        per_disk = [
            parts[i : i + RECORD_SIZE] for i in range(len(parts))[::RECORD_SIZE]
        ]
        return per_disk

    @staticmethod
    def __get_name_and_temp(device_record):
        """
            get device name (without /dev part, to save space on bar)
            and temperature (in °C) as tuple
        """
        device_name = device_record[0].split("/")[-1]
        device_temp = device_record[2]
        return (device_name, device_temp)

    @staticmethod
    def __get_hddtemp(device_record):
        name, temp = device_record
        hddtemp = "{}+{}°C".format(name, temp)
        return hddtemp

    def __get_hddtemps(self):
        data = self.__fetch_data()
        if data is None:
            return "n/a"
        parts = self.__get_parts(data)
        per_disk = self.__partition_parts(parts)
        names_and_temps = [self.__get_name_and_temp(x) for x in per_disk]
        hddtemps = [self.__get_hddtemp(x) for x in names_and_temps]
        return SEPARATOR.join(hddtemps)

    def update(self):
        self.__hddtemps = self.__get_hddtemps()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Test module
"""

import core.widget
import core.module


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config=config, theme=theme, widgets=core.widget.Widget("test"))


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import os
import pytest

import core.config


@pytest.fixture
def defaultConfig():
    return core.config.Config([])


def test_module():
    modules = ["module-1", "module-2", "module-3"]

    cfg = core.config.Config(["-m"] + modules)

    assert cfg.modules() == modules


def test_module_ordering_maintained():
    modules = ["module-1", "module-5", "module-7"]
    more_modules = ["module-0", "module-2", "aaa"]

    cfg = core.config.Config(["-m"] + modules + ["-m"] + more_modules)

    assert cfg.modules() == modules + more_modules


def test_default_interval(defaultConfig):
    assert defaultConfig.interval() == 1


def test_interval():
    interval = 4
    cfg = core.config.Config(["-p", "interval={}".format(interval)])

    assert cfg.interval() == interval


def test_floating_interval():
    interval = 4.5
    cfg = core.config.Config(["-p", "interval={}".format(interval)])

    assert cfg.interval() == interval


def test_default_theme(defaultConfig):
    assert defaultConfig.theme() == "default"


def test_theme():
    theme_name = "sample-theme"
    cfg = core.config.Config(["-t", theme_name])
    assert cfg.theme() == theme_name


def test_default_iconset(defaultConfig):
    assert defaultConfig.iconset() == "auto"


def test_iconset():
    iconset_name = "random-iconset"
    cfg = core.config.Config(["-i", iconset_name])
    assert cfg.iconset() == iconset_name


def test_reverse(defaultConfig):
    assert defaultConfig.reverse() == False

    cfg = core.config.Config(["-r"])

    assert cfg.reverse() == True


def test_logfile(defaultConfig):
    assert defaultConfig.logfile() is None

    logfile = "some-random-logfile"
    cfg = core.config.Config(["-f", logfile])
    assert cfg.logfile() == logfile


def test_all_modules():
    modules = core.config.all_modules()
    assert len(modules) > 0

    for module in modules:
        pyname = "{}.py".format(module)
        base = os.path.abspath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "..",
                "..",
                "bumblebee_status",
                "modules",
            )
        )
    assert os.path.exists(os.path.join(base, "contrib", pyname)) or os.path.exists(
        os.path.join(base, "core", pyname)
    )


def test_list_output(mocker):
    mocker.patch("core.config.sys")
    cfg = core.config.Config(["-l", "themes"])
    cfg = core.config.Config(["-l", "modules"])
    cfg = core.config.Config(["-l", "modules-rst"])


def test_missing_parameter():
    cfg = core.config.Config(["-p", "test.key"])

    assert cfg.get("test.key") == None
    assert cfg.get("test.key", "no-value-set") == "no-value-set"


#
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import core.event


@pytest.fixture
def someEvent():
    class Event:
        def __init__(self):
            core.event.clear()
            self.id = "some event"
            self.called = 0
            self.call_args = []
            self.call_kwargs = []

        def callback(self, *args, **kwargs):
            self.called += 1
            if args:
                self.call_args.append(list(args))
            if kwargs:
                self.call_kwargs.append(kwargs)

    return Event()


def test_simple_callback(someEvent):
    assert someEvent.called == 0

    core.event.register(someEvent.id, someEvent.callback)
    core.event.register(someEvent.id, someEvent.callback)

    core.event.trigger(someEvent.id)

    assert someEvent.called == 2


def test_args_callback(someEvent):
    core.event.register(someEvent.id, someEvent.callback, "a", "b")
    core.event.trigger(someEvent.id)

    assert someEvent.called == 1
    assert len(someEvent.call_args) == 1
    assert someEvent.call_args[0] == ["a", "b"]


def test_kwargs_callback(someEvent):
    core.event.register(
        someEvent.id, someEvent.callback, "a", "b", key1="test", key2="another"
    )
    core.event.trigger(someEvent.id)

    assert someEvent.called == 1
    assert len(someEvent.call_args) == 1
    assert someEvent.call_args[0] == ["a", "b"]
    assert len(someEvent.call_kwargs) == 1
    assert someEvent.call_kwargs[0] == {"key1": "test", "key2": "another"}


def test_arg_trigger(someEvent):
    core.event.register(someEvent.id, someEvent.callback)
    core.event.trigger(someEvent.id, "a", "b")

    assert someEvent.called == 1
    assert len(someEvent.call_args) == 1
    assert someEvent.call_args[0] == ["a", "b"]


def test_kwargs_trigger(someEvent):
    core.event.register(someEvent.id, someEvent.callback)
    core.event.trigger(someEvent.id, "a", "c", key1="test", key2="something")

    assert someEvent.called == 1
    assert len(someEvent.call_args) == 1
    assert someEvent.call_args[0] == ["a", "c"]
    assert len(someEvent.call_kwargs) == 1
    assert someEvent.call_kwargs[0] == {"key1": "test", "key2": "something"}


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays network IO for interfaces.

Parameters:
    * traffic.exclude: Comma-separated list of interface prefixes to exclude (defaults to 'lo,virbr,docker,vboxnet,veth')
    * traffic.states: Comma-separated list of states to show (prefix with '^' to invert - i.e. ^down -> show all devices that are not in state down)
    * traffic.showname: If set to False, hide network interface name (defaults to True)
    * traffic.format: Format string for download/upload speeds.
      Defaults to '{:.2f}'
    * traffic.graphlen: Graph lenth in seconds. Positive even integer. Each
      char shows 2 seconds. If set, enables up/down traffic
      graphs

contributed by `meain <https://github.com/meain>`_ - many thanks!
"""

import re
import time
import psutil
import logging
import netifaces

import core.module

import util.format
import util.graph


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self._exclude = tuple(
            filter(
                len,
                util.format.aslist(
                    self.parameter("exclude", "lo,virbr,docker,vboxnet,veth")
                ),
            )
        )
        self._status = ""

        self._showname = util.format.asbool(self.parameter("showname", True))
        self._format = self.parameter("format", "{:.2f}")
        self._prev = {}
        self._states = {}
        self._lastcheck = 0
        self._states["include"] = []
        self._states["exclude"] = []
        for state in tuple(
            filter(len, util.format.aslist(self.parameter("states", "")))
        ):
            if state[0] == "^":
                self._states["exclude"].append(state[1:])
            else:
                self._states["include"].append(state)
        self._graphlen = int(self.parameter("graphlen", 0))
        if self._graphlen > 0:
            self._graphdata = {}
        self._first_run = True
        self._update_widgets()

    def state(self, widget):
        if "traffic.rx" in widget.name:
            return "rx"
        if "traffic.tx" in widget.name:
            return "tx"
        return self._status

    def update(self):
        try:
            self._update_widgets()
        except Exception as e:
            logging.exception(e)

    def create_widget(self, name, txt=None, attributes={}):
        widget = self.add_widget(name=name, full_text=txt)

        for key in attributes:
            widget.set(key, attributes[key])

        return widget

    def get_addresses(self, intf):
        retval = []
        try:
            for ip in netifaces.ifaddresses(intf).get(netifaces.AF_INET, []):
                if ip.get("addr", "") != "":
                    retval.append(ip.get("addr"))
        except Exception:
            return []
        return retval

    def get_minwidth_str(self):
        """
            computes theme.minwidth string
            based on traffic.format and traffic.graphlen parameters
        """
        minwidth_str = ""
        if self._graphlen > 0:
            graph_len = int(self._graphlen / 2)
            graph_prefix = "0" * graph_len
            minwidth_str += graph_prefix
        minwidth_str += "1000"
        try:
            length = int(re.match(r"{:\.(\d+)f}", self._format).group(1))
            if length > 0:
                minwidth_str += "." + "0" * length
        except AttributeError:
            # return default value
            return "1000.00KiB/s"
        finally:
            minwidth_str += "KiB/s"
        return minwidth_str

    def _update_widgets(self):
        interfaces = [
            i for i in netifaces.interfaces() if not i.startswith(self._exclude)
        ]

        self.clear_widgets()

        counters = psutil.net_io_counters(pernic=True)
        now = time.time()
        timediff = now - (self._lastcheck if self._lastcheck else now)
        if timediff <= 0:
            timediff = 1
        self._lastcheck = now
        for interface in interfaces:
            if self._graphlen > 0:
                if interface not in self._graphdata:
                    self._graphdata[interface] = {
                        "rx": [0] * self._graphlen,
                        "tx": [0] * self._graphlen,
                    }
            if not interface:
                interface = "lo"
            state = "down"
            if len(self.get_addresses(interface)) > 0:
                state = "up"
            elif util.format.asbool(self.parameter("hide_down", True)):
                continue

            if len(self._states["exclude"]) > 0 and state in self._states["exclude"]:
                continue
            if (
                len(self._states["include"]) > 0
                and state not in self._states["include"]
            ):
                continue

            data = {
                "rx": counters[interface].bytes_recv,
                "tx": counters[interface].bytes_sent,
            }

            name = "traffic-{}".format(interface)

            if self._showname:
                self.create_widget(name, interface)

            for direction in ["rx", "tx"]:
                name = "traffic.{}-{}".format(direction, interface)
                widget = self.create_widget(
                    name, attributes={"theme.minwidth": self.get_minwidth_str()},
                )
                prev = self._prev.get(name, 0)
                bspeed = (int(data[direction]) - int(prev)) / timediff
                speed = util.format.byte(bspeed, self._format)
                txtspeed = "{0}/s".format(speed)
                if self._graphlen > 0:
                    # skip first value returned by psutil, because it is
                    # giant and ruins the grapth ratio until it gets pushed
                    # out of saved list
                    if self._first_run is True:
                        self._first_run = False
                    else:
                        self._graphdata[interface][direction] = self._graphdata[
                            interface
                        ][direction][1:]
                        self._graphdata[interface][direction].append(bspeed)
                    txtspeed = "{}{}".format(
                        util.graph.braille(self._graphdata[interface][direction]),
                        txtspeed,
                    )
                widget.full_text(txtspeed)
                self._prev[name] = data[direction]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current date and time.

Parameters:
    * date.format: strftime()-compatible formatting string
    * date.locale: locale to use rather than the system default
"""

import core.decorators
from .datetime import Module


class Module(Module):
    @core.decorators.every(hours=1)
    def __init__(self, config, theme):
        super().__init__(config, theme)

    def default_format(self):
        return "%x"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows Linux kernel version information

contributed by `pierre87 <https://github.com/pierre87>`_ - many thanks!
"""

import platform

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.full_text))

    def full_text(self, widgets):
        return platform.release()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows a widget for each connected screen and allows the user to enable/disable screens.

Parameters:
    * xrandr.overwrite_i3config: If set to 'true', this module assembles a new i3 config
      every time a screen is enabled or disabled by taking the file '~/.config/i3/config.template'
      and appending a file '~/.config/i3/config.<screen name>' for every screen.
    * xrandr.autoupdate: If set to 'false', does *not* invoke xrandr automatically. Instead, the
      module will only refresh when displays are enabled or disabled (defaults to true)

Requires the following python module:
    * (optional) i3 - if present, the need for updating the widget list is auto-detected

Requires the following executable:
    * xrandr
"""

import os
import re
import sys

import core.module
import core.input
import core.decorators

from bumblebee_status.discover import utility

import util.cli
import util.format

try:
    import i3
except:
    pass


class Module(core.module.Module):
    @core.decorators.every(seconds=5)  # takes up to 5s to detect a new screen
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self._autoupdate = util.format.asbool(self.parameter("autoupdate", True))
        self._needs_update = True

        try:
            i3.Subscription(self._output_update, "output")
        except:
            pass

    def _output_update(self, event, data, _):
        self._needs_update = True

    def update(self):
        self.clear_widgets()

        if self._autoupdate == False and self._needs_update == False:
            return

        self._needs_update = False

        for line in util.cli.execute("xrandr -q").split("\n"):
            if not " connected" in line:
                continue
            display = line.split(" ", 2)[0]
            m = re.search(r"\d+x\d+\+(\d+)\+\d+", line)

            widget = self.widget(display)
            if not widget:
                widget = self.add_widget(full_text=display, name=display)
                core.input.register(widget, button=1, cmd=self._toggle)
                core.input.register(widget, button=3, cmd=self._toggle)
            widget.set("state", "on" if m else "off")
            widget.set("pos", int(m.group(1)) if m else sys.maxsize)

        if self._autoupdate == False:
            widget = self.add_widget(full_text="")
            widget.set("state", "refresh")
            core.input.register(widget, button=1, cmd=self._refresh)

    def state(self, widget):
        return widget.get("state", "off")

    def _refresh(self, event):
        self._needs_update = True

    def _toggle(self, event):
        self._refresh(self, event)

        if util.format.asbool(self.parameter("overwrite_i3config", False)) == True:
            toggle_cmd = utility("toggle-display.sh")
        else:
            toggle_cmd = "xrandr"

        widget = self.widget_by_id(event["instance"])

        if widget.get("state") == "on":
            util.cli.execute("{} --output {} --off".format(toggle_cmd, widget.name))
        else:
            first_neighbor = next(
                (widget for widget in self.widgets() if widget.get("state") == "on"),
                None,
            )
            last_neighbor = next(
                (
                    widget
                    for widget in reversed(self.widgets())
                    if widget.get("state") == "on"
                ),
                None,
            )

            neighbor = (
                first_neighbor
                if event["button"] == core.input.LEFT_MOUSE
                else last_neighbor
            )

            if neighbor is None:
                util.cli.execute(
                    "{} --output {} --auto".format(toggle_cmd, widget.name)
                )
            else:
                util.cli.execute(
                    "{} --output {} --auto --{}-of {}".format(
                        toggle_cmd,
                        widget.name,
                        "left"
                        if event.get("button") == core.input.LEFT_MOUSE
                        else "right",
                        neighbor.name,
                    )
                )


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the indicator status, for numlock, scrolllock and capslock 

Parameters:
    * indicator.include: Comma-separated list of interface prefixes to include (defaults to 'numlock,capslock')
    * indicator.signalstype: If you want the signali type color to be 'critical' or 'warning' (defaults to 'warning')

contributed by `freed00m <https://github.com/freed00m>`_ - many thanks!
"""

import core.module

import util.cli
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.__include = tuple(
            filter(
                len, util.format.aslist(self.parameter("include", "NumLock,CapsLock"))
            )
        )
        self.__signalType = (
            self.parameter("signaltype")
            if not self.parameter("signaltype") is None
            else "warning"
        )

    def update(self):
        status_line = ""
        for line in (
            util.cli.execute("xset q", ignore_errors=True).replace(" ", "").split("\n")
        ):
            if "capslock" in line.lower():
                status_line = line
                break
        for indicator in self.__include:
            widget = self.widget(indicator)
            if not widget:
                widget = self.add_widget(name=indicator, full_text=indicator)

            widget.set(
                "status",
                True
                if "{}:on".format(indicator.lower()) in status_line.lower()
                else False,
            )

    def state(self, widget):
        states = []
        if widget.get("status", False):
            states.append(self.__signalType)
        else:
            states.append("normal")
        return states


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""RSS news ticker

Fetches rss news items and shows these as a news ticker.
Left-clicking will open the full story in a browser.
New stories are highlighted.

Parameters:
    * rss.feeds : Space-separated list of RSS URLs
    * rss.length : Maximum length of the module, default is 60

contributed by `lonesomebyte537 <https://github.com/lonesomebyte537>`_ - many thanks!
"""

import feedparser

import webbrowser
import time
import os
import tempfile
import logging
import random
import re
import json

import core.module
import core.widget
import core.input

# pylint: disable=too-many-instance-attributes
class Module(core.module.Module):
    REFRESH_DELAY = 600
    SCROLL_SPEED = 3
    LAYOUT_STYLES_ITEMS = [[1, 1, 1], [3, 3, 2], [2, 3, 3], [3, 2, 3]]
    HISTORY_FILENAME = ".config/i3/rss.hist"

    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.ticker_update))

        self._feeds = self.parameter(
            "feeds", "https://www.espn.com/espn/rss/news"
        ).split(" ")
        self._feeds_to_update = []
        self._response = ""

        self._max_title_length = int(self.parameter("length", 60))

        self._items = []
        self._current_item = None

        self._ticker_offset = 0
        self._pre_delay = 0
        self._post_delay = 0

        self._state = []

        self._newspaper_filename = tempfile.mktemp(".html")

        self._last_refresh = 0
        self._last_update = 0

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self._open)
        core.input.register(
            self, button=core.input.RIGHT_MOUSE, cmd=self._create_newspaper
        )

        self._history = {"ticker": {}, "newspaper": {}}
        self._load_history()

    def _load_history(self):
        if os.path.isfile(self.HISTORY_FILENAME):
            self._history = json.loads(open(self.HISTORY_FILENAME, "r").read())

    def _update_history(self, group):
        sources = set([i["source"] for i in self._items])
        self._history[group] = dict(
            [
                [s, [i["title"] for i in self._items if i["source"] == s]]
                for s in sources
            ]
        )

    def _save_history(self):
        if not os.path.exists(os.path.dirname(self.HISTORY_FILENAME)):
            os.makedirs(os.path.dirname(self.HISTORY_FILENAME))
        open(self.HISTORY_FILENAME, "w").write(json.dumps(self._history))

    def _check_history(self, items, group):
        for i in items:
            i["new"] = not (
                i["source"] in self._history[group]
                and i["title"] in self._history[group][i["source"]]
            )

    def _open(self, _):
        if self._current_item:
            webbrowser.open(self._current_item["link"])

    def _check_for_image(self, entry):
        image = next(
            iter([l["href"] for l in entry["links"] if l["rel"] == "enclosure"]), None
        )
        if not image and "media_content" in entry:
            try:
                media = sorted(
                    entry["media_content"],
                    key=lambda i: i["height"] if "height" in i else 0,
                    reverse=True,
                )
                image = next(
                    iter([i["url"] for i in media if i["medium"] == "image"]), None
                )
            except Exception:
                pass
        if not image:
            match = re.search(
                r"<img[^>]*src\s*=['\']*([^\s^>^'^\']*)['\']*", entry["summary"]
            )
            if match:
                image = match.group(1)
        return image if image else ""

    def _remove_tags(self, txt):
        return re.sub(r"<[^>]*>", "", txt)

    def _create_item(self, entry, url, feed):
        return {
            "title": self._remove_tags(entry["title"].replace("\n", " ")),
            "link": entry["link"],
            "new": True,
            "source": url,
            "summary": self._remove_tags(entry["summary"]),
            "feed": feed,
            "image": self._check_for_image(entry),
            "published": time.mktime(entry.published_parsed)
            if hasattr(entry, "published_parsed")
            else 0,
        }

    def _update_items_from_feed(self, url):
        parser = feedparser.parse(url)
        new_items = [
            self._create_item(entry, url, parser["feed"]["title"])
            for entry in parser["entries"]
        ]
        # Check history
        self._check_history(new_items, "ticker")
        # Remove the previous items
        self._items = [i for i in self._items if i["source"] != url]
        # Add the new items
        self._items.extend(new_items)
        # Sort the items on publish date
        self._items.sort(key=lambda i: i["published"], reverse=True)

    def _check_for_refresh(self):
        if self._feeds_to_update:
            # Update one feed at a time to not overload this update cycle
            url = self._feeds_to_update.pop()
            self._update_items_from_feed(url)

            if not self._feeds_to_update:
                self._update_history("ticker")
                self._save_history()

            if not self._current_item:
                self._next_item()
        elif time.time() - self._last_refresh >= self.REFRESH_DELAY:
            # Populate the list with feeds to update
            self._feeds_to_update = self._feeds[:]
            # Update the refresh time
            self._last_refresh = time.time()

    def _next_item(self):
        self._ticker_offset = 0
        self._pre_delay = 2
        self._post_delay = 4

        if not self._items:
            return

        # Index of the current element
        idx = (
            self._items.index(self._current_item)
            if self._current_item in self._items
            else -1
        )

        # First show new items, else show next
        new_items = [i for i in self._items if i["new"]]
        self._current_item = next(
            iter(new_items), self._items[(idx + 1) % len(self._items)]
        )

    def _check_scroll_done(self):
        # Check if the complete title has been shown
        if self._ticker_offset + self._max_title_length > len(
            self._current_item["title"]
        ):
            # Do not immediately show next item after scroll
            self._post_delay -= 1
            if self._post_delay == 0:
                self._current_item["new"] = False
                # Mark the previous item as 'old'
                self._next_item()
        else:
            # Increase scroll position
            self._ticker_offset += self.SCROLL_SPEED

    def ticker_update(self, _):
        # Only update the ticker once a second
        now = time.time()
        if now - self._last_update < 1:
            return self._response

        self._last_update = now

        self._check_for_refresh()

        # If no items were retrieved, return an empty string
        if not self._current_item:
            return " " * self._max_title_length

        # Prepare a substring of the item title
        self._response = self._current_item["title"][
            self._ticker_offset : self._ticker_offset + self._max_title_length
        ]
        # Add spaces if too short
        self._response = self._response.ljust(self._max_title_length)

        # Do not immediately scroll
        if self._pre_delay > 0:
            # Change state during pre_delay for new items
            if self._current_item["new"]:
                self._state = ["warning"]
            self._pre_delay -= 1
            return self._response

        self._state = []
        self._check_scroll_done()

        return self._response

    def state(self, _):
        return self._state

    def _create_news_element(self, item, overlay_title):
        try:
            timestr = (
                "" if item["published"] == 0 else str(time.ctime(item["published"]))
            )
        except Exception as exc:
            logging.error(str(exc))
            raise e
        element = "<div class='item' onclick=window.open('" + item["link"] + "')>"
        element += "<div class='titlecontainer'>"
        element += (
            "  <img "
            + ("" if item["image"] else "class='noimg' ")
            + "src='"
            + item["image"]
            + "'>"
        )
        element += (
            "  <div class='title"
            + (" overlay" if overlay_title else "")
            + "'>"
            + ("<span class='star'>&#x2605;</span>" if item["new"] else "")
            + item["title"]
            + "</div>"
        )
        element += "</div>"
        element += "<div class='summary'>" + item["summary"] + "</div>"
        element += (
            "<div class='info'><span class='author'>"
            + item["feed"]
            + "</span><span class='published'>"
            + timestr
            + "</span></div>"
        )
        element += "</div>"
        return element

    def _create_news_section(self, newspaper_items):
        style = random.randint(0, 3)
        section = "<table><tr class='style" + str(style) + "'>"
        for i in range(0, 3):
            section += "<td><div class='itemcontainer'>"
            for _ in range(0, self.LAYOUT_STYLES_ITEMS[style][i]):
                if newspaper_items:
                    section += self._create_news_element(
                        newspaper_items[0], self.LAYOUT_STYLES_ITEMS[style][i] != 3
                    )
                    del newspaper_items[0]
            section += "</div></td>"
        section += "</tr></table>"
        return section

    def _create_newspaper(self, _):
        content = ""
        newspaper_items = self._items[:]
        self._check_history(newspaper_items, "newspaper")

        # Make sure new items are always listed first, independent of publish date
        newspaper_items.sort(
            key=lambda i: i["published"] + (10000000 if i["new"] else 0), reverse=True
        )

        while newspaper_items:
            content += self._create_news_section(newspaper_items)
        open(self._newspaper_filename, "w").write(
            HTML_TEMPLATE.replace("[[CONTENT]]", content)
        )
        webbrowser.open("file://" + self._newspaper_filename)
        self._update_history("newspaper")
        self._save_history()


HTML_TEMPLATE = """<!DOCTYPE html>
<html>
<head>
<script>
window.onload = function() {
    var images = document.getElementsByTagName('img'); 
    // Remove very small images
    for(var i = 0; i < images.length; i++) {
        if (images[i].naturalWidth<50 || images[i].naturalHeight<50) {
            images[i].src = ''
            images[i].className+=' noimg'
        }
    }
}
</script>
</head>
<style>
    body {background: #eee; font-family: Helvetica neue;}
    td {background: #fff; height: 100%;}
    tr.style0 td {width: 33%;}
    tr.style1 td {width: 20%;}
    tr.style1 td:last-child {width: 60%;}
    tr.style2 td {width: 20%;}
    tr.style2 td:first-child {width: 60%;}
    tr.style3 td {width: 20%;}
    tr.style3 td:nth-child(2) {width: 60%;}
    img {width: 100%; display: block; }
    img.noimg {min-height:250px; background: #1299c8;}
    #content {width: 1500px; margin: auto; background: #eee; padding: 1px;}
    #newspapertitle {text-align: center; font-size: 60px; font-family: Arial Black; background: #1299c8; font-style: Italic; padding: 10px; color: #fff; }
    .star {color: #ffa515; font-size: 24px;}
    .section {display: flex;}
    .column {display: flex;}
    .itemcontainer {width: 100%; height: 100%; position: relative; display: inline-table;}
    .item {cursor: pointer; }
    .titlecontainer {position: relative;}
    .title.overlay {font-family: Arial; position: absolute; bottom: 10px; color: #fff; font-weight: bold; text-align: right; max-width: 75%; right: 10px; font-size: 23px; text-shadow: 1px 0 0 #000, 0 -1px 0 #000, 0 1px 0 #000, -1px 0 0 #000;}
    .title:not(.overlay) {font-weight: bold; padding: 0px 10px;}
    .summary {color: #444; padding: 10px 10px 0px 10px; font-family: Times new roman; font-size: 18px; flex: 1;max-height: 105px; overflow: hidden;}
    .info {color: #aaa; font-family: arial; font-size: 13px; padding: 10px;}
    .published {float: right;}
</style>
<body>
  <div id='content'>
  <div id='newspapertitle'>Bumblebee Daily</div>
    [[CONTENT]]
  </div>
</body>
</html>"""
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays CPU utilization across all CPUs.

Requirements:
    * the psutil Python module for the first three items from the list above

Parameters:
    * cpu.warning : Warning threshold in % of CPU usage (defaults to 70%)
    * cpu.critical: Critical threshold in % of CPU usage (defaults to 80%)
    * cpu.format  : Format string (defaults to '{:.01f}%')
"""

import psutil

import core.module
import core.widget
import core.input


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.utilization))
        self.widget().set("theme.minwidth", self._format.format(100.0 - 10e-20))
        self._utilization = psutil.cpu_percent(percpu=False)
        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="gnome-system-monitor"
        )

    @property
    def _format(self):
        return self.parameter("format", "{:.01f}%")

    def utilization(self, _):
        return self._format.format(self._utilization)

    def update(self):
        self._utilization = psutil.cpu_percent(percpu=False)

    def state(self, _):
        return self.threshold_state(self._utilization, 70, 80)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: utf-8 -*-

import re


def asbool(val):
    """Converts a value into a boolean

    :param val: value to convert; accepts a wide range of
        possible representations, such as yes, no, true, false, on, off

    :return: True of val maps to true, False otherwise
    :rtype: boolean
    """
    if val is None:
        return False
    if isinstance(val, bool):
        return val
    val = str(val).strip().lower()
    return val in ("t", "true", "y", "yes", "on", "1")


def asint(val, minimum=None, maximum=None):
    """Converts a value into an integer

    :param val: value to convert
    :param minimum: if specified, this determines the lower
        boundary for the returned value, defaults to None
    :param maximum: if specified, this determines the upper
        boundary for the returned value, defaults to None

    :return: integer representation of value
    :rtype: integer
    """
    if val is None:
        val = 0
    val = int(val)
    val = min(val, maximum if maximum else val)
    val = max(val, minimum if minimum else val)
    return val


def aslist(val):
    """Converts a comma-separated value string into a list

    :param val: value to convert, either a single value or a comma-separated string

    :return: list representation of the value passed in
    :rtype: list of strings
    """
    if val is None:
        return []
    if isinstance(val, list):
        return val
    return str(val).replace(" ", "").split(",")


__UNITS = {"metric": "C", "kelvin": "K", "imperial": "F", "default": "C"}


def astemperature(val, unit="metric"):
    """Returns a temperature representation of the input value

    :param val: value to format, must be convertible into an integer
    :param unit: unit of the input value, supported units are:
        metric, kelvin, imperial, defaults to metric

    :return: temperature representation of the input value
    :rtype: string
    """
    return "{}°{}".format(int(val), __UNITS.get(unit.lower(), __UNITS["default"]))


def byte(val, fmt="{:.2f}"):
    """Returns a byte representation of the input value

    :param val: value to format, must be convertible into a float
    :param fmt: optional output format string, defaults to {:.2f}

    :return: byte representation (e.g. <X> KiB, GiB, etc.) of the input value
    :rtype: string
    """

    val = float(val)
    for unit in ["", "Ki", "Mi", "Gi"]:
        if val < 1024.0:
            return "{}{}B".format(fmt, unit).format(val)
        val /= 1024.0
    return "{}GiB".format(fmt).format(val * 1024.0)


__seconds_pattern = re.compile(r"(([\d\.?]+)h)?(([\d\.]+)m)?([\d\.]+)?s?")


def seconds(duration):
    """Returns a time duration (in seconds) representation of the input value

    :param duration: value to format (e.g. 5h30m2s)

    :return: duration in seconds of the input value
    :rtype: float
    """
    if isinstance(duration, int) or isinstance(duration, float):
        return float(duration)

    matches = __seconds_pattern.match(duration)
    result = 0.0
    if matches.group(2):
        result += float(matches.group(2)) * 3600  # hours
    if matches.group(4):
        result += float(matches.group(4)) * 60  # minutes
    if matches.group(5):
        result += float(matches.group(5))  # seconds

    return result


def duration(duration, compact=False, unit=False):
    """Returns a time duration string representing the input value

    :param duration: value to format, must be convertible into an into
    :param compact: whether to show also seconds, defaults to False
    :param unit: whether to display he unit, defaults to False

    :return: duration representation (e.g. 5:02s) of the input value
    :rtype: string
    """
    duration = int(duration)
    if duration < 0:
        return "n/a"
    minutes, seconds = divmod(duration, 60)
    hours, minutes = divmod(minutes, 60)
    suf = "m"
    res = "{:02d}:{:02d}".format(minutes, seconds)
    if hours > 0:
        if compact:
            res = "{:02d}:{:02d}".format(hours, minutes)
        else:
            res = "{:02d}:{}".format(hours, res)
        suf = "h"

    return "{}{}".format(res, suf if unit else "")


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
